#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "svm.h"

#include "mex.h"

#if MX_API_VER < 0x07030000
typedef int mwIndex;
#endif 

#define CMD_LEN 2048
#define Malloc(type,n) (type *)mxCalloc((n),sizeof(type))
#define Free(v) mxFree(v)

void print_null(const char *s) {}
void print_string_matlab(const char *s) {mexPrintf(s); mexEvalString("drawnow;");}
const char *svm_check_parameter(const struct svm_problem *prob, const struct svm_parameter *param);

void exit_with_help()
{
	mexPrintf(
	"QP solver by SMO\n"
	"      min 0.5 alpha'*(K.*(y*y'))*alpha + b'*alpha\n"
	"       s.t. 0<=alpha<=C, y'*alpha=d (=y'*alpha0)\n\n"
	" [alpha status] = smo(y, K, b, C, alpha0, params);\n\n"
	"Input:\n"
	"  y - Label vector, usually {-1,+1} [n x 1]\n"
	"  K - Gram matrix  [n x n]\n"
	"  b - Bias vector  [n x 1]\n"
	"  C - Upper bounds [n x 1]\n"
	"      If C = [], the C values are set as indicated by params.c.\n"
	"  alpha0 - Initial values [n x 1]\n"
	"  params - Parameter struct structure\n"
	"             .c cost : set the parameter C (default 1)\n"
	"             .e epsilon : set tolerance of termination criterion (default 0.001)\n"
	"             .i iteration: set maximum number of iteration (default Inf)\n"
	"             .q 0/1: set silence, 1 for silent, 0 for verbose (default 0)\n"
	"             .m cachesize : set cache memory size in MB (default 100)\n"
	"             .h shrinking: whether to use the shrinking heuristics, 0 or 1 (default 1)\n"
	"           Or\n"
	"           Parameter string, such as '-c 1 -m 100 -e 0.001 -i 10000000 -h 1'\n"
	"Output:\n"
	" alpha - Solutoin [n x 1]\n"
	" status- Optimization status structure\n"
	"           .cost - Cost value\n"
	"           .rho  - Bias value; classifier by (y.*alpha)'K - rho\n"
	"           .cond - Condition, 1:optimized, 0:not optimized until max_iter\n"
	"           .iter - Number of iteration in optimization\n"
	);
}

// svm arguments
struct svm_parameter param;		// set by parse_command_line
struct svm_problem prob;		// set by read_problem
schar  *y;
double *b;
double *C;
double *alpha0;

// nrhs should be 6
int parse_command_line(int nrhs, const mxArray *prhs[])
{
	const mxArray *params;
	int quiet = 0;
	void (*print_func)(const char *) = print_string_matlab;	// default printing to matlab display

	// default values
	param.cache_size = 100;
	param.C = 1;
	param.eps = 1e-3;
	param.shrinking = 1;
	param.maxiter = (int)INT_MAX;

	if(nrhs <= 4) return 1;
	if(nrhs == 5) return 0;

	params = prhs[5];

	if(mxIsStruct(params)){
		mxArray *tmp_field; 
		mwIndex index_params = 0;

		// parse options
		int nField = (int)mxGetNumberOfFields(params);
		for(int i = 0; i < nField; i++){
			const char* fieldname = mxGetFieldNameByNumber(params, i);
			if(strcmp(fieldname,"m")==0){
				tmp_field        = mxGetField(params, index_params, "m"); 
				param.cache_size = (double)*mxGetPr(tmp_field);
			}else if(strcmp(fieldname,"c")==0){
				tmp_field = mxGetField(params, index_params, "c"); 
				param.C   = (double)*mxGetPr(tmp_field);
			}else if(strcmp(fieldname,"e")==0){
				tmp_field = mxGetField(params, index_params, "e"); 
				param.eps = (double)*mxGetPr(tmp_field);
			}else if(strcmp(fieldname,"h")==0){
				tmp_field       = mxGetField(params, index_params, "h"); 
				param.shrinking = (int)*mxGetPr(tmp_field);
			}else if(strcmp(fieldname,"i")==0){
				tmp_field     = mxGetField(params, index_params, "i"); 
				param.maxiter = (int)*mxGetPr(tmp_field);
			}else if(strcmp(fieldname,"q")==0){
				tmp_field = mxGetField(params, index_params, "q"); 
				quiet     = (int)*mxGetPr(tmp_field);
				if(quiet) print_func = &print_null;
			}else{
				mexPrintf("Error: unknown parameter option\n");
				return 1;
			}
		}
	}else if(mxIsChar(params)){
		int i, argc = 1;
		char cmd[CMD_LEN];
		char *argv[CMD_LEN/2];

		// put options in argv[]
		mxGetString(params, cmd,  (int)mxGetN(params) + 1);
		if((argv[argc] = strtok(cmd, " ")) == NULL)
			return 0;
		while((argv[++argc] = strtok(NULL, " ")) != NULL)
			;

		// parse options
		for(i=1;i<argc;i++) {
			if(argv[i][0] != '-') break;
			if(++i>=argc) return 1;
			switch(argv[i-1][1]) {
				case 'm':
					param.cache_size = atof(argv[i]);
					break;
				case 'c':
					param.C = atof(argv[i]);
					break;
				case 'e':
					param.eps = atof(argv[i]);
					break;
				case 'h':
					param.shrinking = atoi(argv[i]);
					break;
				case 'i':
					param.maxiter = atoi(argv[i]);
					break;
				case 'q':
					quiet = atoi(argv[i]);
					if(quiet) print_func = &print_null;
					break;
				default:
					mexPrintf("Error: unknown parameter option\n");
					return 1;
			}
		}
	}else{
		mexPrintf("Error: cannot parse parameters\n");
	   	return 1;
	}

	svm_set_print_string_function(print_func);

	return 0;
}

// read in a problem (in svmlight format)
int read_problem_dense(const mxArray *label_vec, const mxArray *instance_mat, 
		const mxArray *b_mat, const mxArray *c_mat,const mxArray *initalpha_mat)
{
	int i, j;
	int sc;
	double *samples, *y_;

	// the number of instance
	prob.l = (int)mxGetM(instance_mat);

	prob.x  = (double*)mxGetPr(instance_mat);
	y       = Malloc(schar,  prob.l);
	C       = Malloc(double, prob.l);

	// y
	if(mxGetNumberOfElements(label_vec) != prob.l){ 
		mexPrintf("Error: wrong size of label\n");
		return -1;
	}else y_ = (double*)mxGetPr(label_vec);
	for(i = 0; i < prob.l; i++)
		if(y_[i] > 0) y[i] = 1;
		else          y[i] = -1;

	// b
	if(mxGetNumberOfElements(b_mat) != prob.l){ 
		mexPrintf("Error: wrong size of b\n"); 
		return -1;
	}else b = (double*)mxGetPr(b_mat);

	// C
	if(mxIsEmpty(c_mat)){
		for(int ci = 0; ci < prob.l; ci++) C[ci] = param.C;
	}else if(mxGetNumberOfElements(c_mat) != prob.l){
		mexPrintf("Error: wrong size of C\n"); 
	   	return -1;
	}else memcpy(C,(double*)mxGetPr(c_mat),sizeof(double)*prob.l);

	// alpha0
	if(mxGetNumberOfElements(initalpha_mat) != prob.l){
		mexPrintf("Error: wrong size of alpha0\n"); 
	   	return -1;
	}else alpha0 = (double*)mxGetPr(initalpha_mat);

	return 0;
}

static void fake_answer(mxArray *plhs[])
{
	plhs[0] = mxCreateDoubleMatrix(0, 0, mxREAL);
}

// Interface function of matlab
// now assume prhs[0]: label prhs[1]: features
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	const char *error_msg;

	// fix random seed to have same results for each run
	// (for cross validation and probability estimation)
	srand(1);

	// Translate the input Matrix to the format such that svmtrain.exe can recognize it
	if(nrhs > 0 && nrhs < 7)
	{
		int err;

		if(parse_command_line(nrhs, prhs))
		{
			fake_answer(plhs);
			return;
		}

		if(mxIsSparse(prhs[1]))
		{
			mexPrintf("Error: not support sparce matrix\n");
			fake_answer(plhs);
			return;
		}
		else
			err = read_problem_dense(prhs[0], prhs[1], prhs[2], prhs[3], prhs[4]);

		// svmtrain's original code
		error_msg = svm_check_parameter(&prob, &param);

		if(err || error_msg)
		{
			if (error_msg != NULL)
				mexPrintf("Error: %s\n", error_msg);
			Free(y);
			Free(C);
			fake_answer(plhs);
			return;
		}

		plhs[0] = mxCreateDoubleMatrix(prob.l, 1, mxREAL);
		double* alpha = mxGetPr(plhs[0]);
		memcpy(alpha, alpha0, sizeof(double)*prob.l);

		struct decision_function model;
	    model = solveSMO(&prob, &param, y, b, C, alpha);

		if(nlhs > 1){
			static const char *fnames[] = { "cost", "rho", "cond", "iter" };
			plhs[1] = mxCreateStructMatrix(1, 1, 4, fnames);
			mxArray *fout;
		   	fout = mxCreateDoubleScalar(model.obj);
			mxSetFieldByNumber(plhs[1], 0, 0, fout);
		   	fout = mxCreateDoubleScalar(model.rho);
			mxSetFieldByNumber(plhs[1], 0, 1, fout);
		   	fout = mxCreateDoubleScalar(model.chk);
			mxSetFieldByNumber(plhs[1], 0, 2, fout);
		   	fout = mxCreateDoubleScalar(model.iter);
			mxSetFieldByNumber(plhs[1], 0, 3, fout);
		}

		Free(y);
		Free(C);
	}
	else
	{
		exit_with_help();
		fake_answer(plhs);
	}
	return;
}

const char *svm_check_parameter(const struct svm_problem *prob, const struct svm_parameter *param)
{
	// cache_size,eps,C,nu,p,shrinking

	if(param->cache_size <= 0) return "cache_size <= 0";

	if(param->eps <= 0) return "eps <= 0";

	if(param->C <= 0) return "C <= 0";

	if(param->shrinking != 0 &&
	   param->shrinking != 1)
		return "shrinking != 0 and shrinking != 1";

	return NULL;
}
